import warnings
from contextlib import contextmanager
from unittest import TestCase

try:
    from collections import Counter
except ImportError:
    from counter import Counter

try:
    from unittest import skip, skipUnless
except ImportError:
    def skip(_f):
        return lambda self: None


    def skipUnless(condition, _reason):
        if condition:
            return lambda x: x
        else:
            return lambda x: None


@contextmanager
def temp_attrs(*attrs):
    if len(attrs) == 3 and not isinstance(attrs[1], tuple):
        attrs = (attrs,)
    previous = []
    try:
        for obj, attr, val in attrs:
            old = getattr(obj, attr)
            previous.append((obj, attr, old))
            setattr(obj, attr, val)
        yield
    finally:
        for t in previous:
            setattr(*t)


class TestCaseWithUtils(TestCase):
    def assert_warns(self, category, message, f):
        with warnings.catch_warnings(record=True) as warning_list:
            warnings.simplefilter('always')
            result = f()
            self.assertEqual(len(warning_list), 1)
            self.assertEqual(warning_list[0].category, category)
            self.assertEqual(str(warning_list[0].message), message)
            return result


def assert_unique(items):
    counts = Counter(items)
    dups = [k for k, v in counts.items()
            if v > 1]
    if dups:
        raise ValueError('Duplicates: %s' % dups)


class OldStyleClass:
    pass



import os
import re
import sys
import unittest
from array import array
from collections import defaultdict, deque
from sys import version_info, version
from unittest import skipIf

from tests.utils import TestCaseWithUtils, temp_attrs, assert_unique, Counter, skipUnless, OldStyleClass

try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict

try:
    from collections import ChainMap
except ImportError:
    from chainmap import ChainMap

# Python 3.9 compatibility (importing Set from collections is deprecated)
if version_info.major == 2:
    from collections import Set
else:
    from collections.abc import Set

from cheap_repr import basic_repr, register_repr, cheap_repr, PY2, PY3, ReprSuppressedWarning, find_repr_function, \
    raise_exceptions_from_default_repr, repr_registry

PYPY = 'pypy' in version.lower()


class FakeExpensiveReprClass(object):
    def __repr__(self):
        return 'bad'


register_repr(FakeExpensiveReprClass)(basic_repr)


class ErrorClass(object):
    def __init__(self, error=False):
        self.error = error

    def __repr__(self):
        if self.error:
            raise ValueError()
        return 'bob'


class ErrorClassChild(ErrorClass):
    pass


class OldStyleErrorClass:
    def __init__(self, error=False):
        self.error = error

    def __repr__(self):
        if self.error:
            raise ValueError()
        return 'bob'


class OldStyleErrorClassChild(OldStyleErrorClass):
    pass


class DirectRepr(object):
    def __init__(self, r):
        self.r = r

    def __repr__(self):
        return self.r


class RangeSet(Set):
    def __init__(self, length):
        self.length = length

    def __contains__(self, x):
        pass

    def __iter__(self):
        for x in range(self.length):
            yield x

    def __len__(self):
        return self.length


class NormalClass(object):
    def __init__(self, x):
        self.x = x

    def __repr__(self):
        return repr(self.x)

    def foo(self):
        pass


class TestCheapRepr(TestCaseWithUtils):
    maxDiff = None

    def normalise_repr(self, string):
        string = re.sub(r'0x[0-9a-f]+', '0xXXX', string)
        string = re.sub('\\s+\n', '\n', string)
        string = re.sub('\n\n', '\n', string)
        return string

    def assert_cheap_repr(self, x, expected_repr):
        actual = self.normalise_repr(cheap_repr(x))
        self.assertEqual(actual, expected_repr)

    def assert_usual_repr(self, x, normalise=False):
        expected = repr(x)
        if normalise:
            expected = self.normalise_repr(expected)
        self.assert_cheap_repr(x, expected)

    def assert_cheap_repr_evals(self, s):
        self.assert_cheap_repr(eval(s), s)

    def assert_cheap_repr_warns(self, x, message, expected_repr):
        self.assert_warns(ReprSuppressedWarning,
                          message,
                          lambda: self.assert_cheap_repr(x, expected_repr))

    def test_registered_default_repr(self):
        x = FakeExpensiveReprClass()
        self.assertEqual(repr(x), 'bad')
        self.assert_cheap_repr(x, r'<FakeExpensiveReprClass instance at 0xXXX>')

    def test_bound_method(self):
        self.assert_usual_repr(NormalClass('hello').foo)
        self.assert_cheap_repr(
            RangeSet(10).__len__,
            '<bound method RangeSet.__len__ of RangeSet({0, 1, 2, 3, 4, 5, ...})>')

    def test_chain_map(self):
        self.assert_usual_repr(ChainMap({1: 2, 3: 4}, dict.fromkeys('abcd')))

        ex = (
            "ChainMap("
            "OrderedDict([('1', 0), ('2', 0), ('3', 0), ('4', 0), ...]), "
            "OrderedDict([('1', 0), ('2', 0), ('3', 0), ('4', 0), ...]), "
            "..., "
            "OrderedDict([('1', 0), ('2', 0), ('3', 0), ('4', 0), ...]), "
            "OrderedDict([('1', 0), ('2', 0), ('3', 0), ('4', 0), ...])"
            ")"
        ) if sys.version_info < (3, 12) else (
            "ChainMap("
            "OrderedDict({'1': 0, '2': 0, '3': 0, '4': 0, '5': 0, ...}), "
            "OrderedDict({'1': 0, '2': 0, '3': 0, '4': 0, '5': 0, ...}), "
            "..., "
            "OrderedDict({'1': 0, '2': 0, '3': 0, '4': 0, '5': 0, ...}), "
            "OrderedDict({'1': 0, '2': 0, '3': 0, '4': 0, '5': 0, ...})"
            ")"
        )
        self.assert_cheap_repr(ChainMap(*[OrderedDict.fromkeys('1234567890', 0) for _ in range(10)]),
                               ex)

    def test_list(self):
        self.assert_usual_repr([])
        self.assert_usual_repr([1, 2, 3])
        self.assert_cheap_repr([1, 2, 3] * 10 + [4, 5, 6, 7], '[1, 2, 3, ..., 5, 6, 7]')

    def test_tuple(self):
        self.assert_usual_repr(())
        self.assert_usual_repr((1,))
        self.assert_usual_repr((1, 2, 3))
        self.assert_cheap_repr((1, 2, 3) * 10 + (4, 5, 6, 7), '(1, 2, 3, ..., 5, 6, 7)')

    def test_sets(self):
        self.assert_usual_repr(set())
        self.assert_usual_repr(frozenset())
        self.assert_usual_repr(set([1, 2, 3]))
        self.assert_usual_repr(frozenset([1, 2, 3]))
        self.assert_cheap_repr(set(range(10)),
                               'set([0, 1, 2, 3, 4, 5, ...])' if PY2 else
                               '{0, 1, 2, 3, 4, 5, ...}')

    def test_dict(self):
        self.assert_usual_repr({})
        d1 = {1: 2, 2: 3, 3: 4}
        self.assert_usual_repr(d1)
        d2 = dict((x, x * 2) for x in range(10))
        self.assert_cheap_repr(d2, '{0: 0, 1: 2, 2: 4, 3: 6, ...}')
        self.assert_cheap_repr(
            {'a' * 100: 'b' * 100},
            "{'aaaaaaaaaaaaaaaaaaaaaaaaaaaa...aaaaaaaaaaaaaaaaaaaaaaaaaaaaa': 'bbbbbbbbbbbbbbbbbbbbbbbbbbbb...bbbbbbbbbbbbbbbbbbbbbbbbbbbbb'}")

        if PY3:
            self.assert_usual_repr({}.keys())
            self.assert_usual_repr({}.values())
            self.assert_usual_repr({}.items())

            self.assert_usual_repr(d1.keys())
            self.assert_usual_repr(d1.values())
            self.assert_usual_repr(d1.items())

            self.assert_cheap_repr(d2.keys(),
                                   'dict_keys([0, 1, 2, 3, 4, 5, ...])')
            self.assert_cheap_repr(d2.values(),
                                   'dict_values([0, 2, 4, 6, 8, 10, ...])')
            self.assert_cheap_repr(d2.items(),
                                   'dict_items([(0, 0), (1, 2), (2, 4), (3, 6), ...])')

    def test_defaultdict(self):
        d = defaultdict(int)
        self.assert_usual_repr(d)
        d.update({1: 2, 2: 3, 3: 4})
        self.assert_usual_repr(d)
        d.update(dict((x, x * 2) for x in range(10)))
        self.assertTrue(cheap_repr(d) in
                        ("defaultdict(%r, {0: 0, 1: 2, 2: 4, 3: 6, ...})" % int,
                         "defaultdict(%r, {1: 2, 2: 4, 3: 6, 0: 0, ...})" % int))

    def test_deque(self):
        self.assert_usual_repr(deque())
        self.assert_usual_repr(deque([1, 2, 3]))
        self.assert_cheap_repr(deque(range(10)), 'deque([0, 1, 2, 3, 4, 5, ...])')

    def test_ordered_dict(self):
        self.assert_usual_repr(OrderedDict())
        self.assert_usual_repr(OrderedDict((x, x * 2) for x in range(3)))
        self.assert_cheap_repr(OrderedDict((x, x * 2) for x in range(10)),
                               'OrderedDict([(0, 0), (1, 2), (2, 4), (3, 6), ...])'
                               if sys.version_info < (3, 12) else
                               'OrderedDict({0: 0, 1: 2, 2: 4, 3: 6, 4: 8, ...})')

    def test_counter(self):
        self.assert_usual_repr(Counter())
        self.assert_cheap_repr_evals('Counter({0: 0, 2: 1, 4: 2})')
        self.assert_cheap_repr(Counter(dict((x * 2, x) for x in range(10))),
                               'Counter(10 keys)')

    def test_array(self):
        self.assert_usual_repr(array('l', []))
        self.assert_usual_repr(array('l', [1, 2, 3, 4, 5]))
        self.assert_cheap_repr(array('l', range(10)),
                               "array('l', [0, 1, 2, ..., 8, 9])")

    def test_django_queryset(self):
        os.environ['DJANGO_SETTINGS_MODULE'] = 'tests.fake_django_settings'
        import django

        django.setup()
        from django.contrib.contenttypes.models import ContentType
        self.assert_cheap_repr(ContentType.objects.all(),
                               '<QuerySet instance of ContentType at 0xXXX>')

    if not PYPY and version_info[:2] < (3, 8):
        def test_numpy_array(self):
            import numpy

            self.assert_usual_repr(numpy.array([]))
            self.assert_usual_repr(numpy.array([1, 2, 3, 4, 5]))
            self.assert_cheap_repr(numpy.array(range(10)),
                                   'array([0, 1, 2, ..., 7, 8, 9])')

            self.assert_cheap_repr(numpy.arange(100).reshape(10, 10),
                                   """\
array([[ 0,  1,  2, ...,  7,  8,  9],
       [10, 11, 12, ..., 17, 18, 19],
       [20, 21, 22, ..., 27, 28, 29],
       ...,
       [70, 71, 72, ..., 77, 78, 79],
       [80, 81, 82, ..., 87, 88, 89],
       [90, 91, 92, ..., 97, 98, 99]])""")

            self.assert_cheap_repr(numpy.arange(1000).reshape(10, 10, 10),
                                   """\
array([[[  0,   1, ...,   8,   9],
        [ 10,  11, ...,  18,  19],
        ...,
        [ 80,  81, ...,  88,  89],
        [ 90,  91, ...,  98,  99]],
       [[100, 101, ..., 108, 109],
        [110, 111, ..., 118, 119],
        ...,
        [180, 181, ..., 188, 189],
        [190, 191, ..., 198, 199]],
       ...,
       [[800, 801, ..., 808, 809],
        [810, 811, ..., 818, 819],
        ...,
        [880, 881, ..., 888, 889],
        [890, 891, ..., 898, 899]],
       [[900, 901, ..., 908, 909],
        [910, 911, ..., 918, 919],
        ...,
        [980, 981, ..., 988, 989],
        [990, 991, ..., 998, 999]]])""")

            self.assert_cheap_repr(numpy.arange(10000).reshape(10, 10, 10, 10),
                                   """\
array([[[[   0, ...,    9],
         ...,
         [  90, ...,   99]],
        ...,
        [[ 900, ...,  909],
         ...,
         [ 990, ...,  999]]],
       ...,
       [[[9000, ..., 9009],
         ...,
         [9090, ..., 9099]],
        ...,
        [[9900, ..., 9909],
         ...,
         [9990, ..., 9999]]]])""")

            self.assert_cheap_repr(numpy.arange(128).reshape(2, 2, 2, 2, 2, 2, 2),
                                   "array(dtype('int64'), shape=(2, 2, 2, 2, 2, 2, 2))")

            self.assert_cheap_repr(numpy.ma.array([1, 2, 3], mask=[0, 1, 0]),
                                   "MaskedArray(dtype('int64'), shape=(3,))")

            self.assert_cheap_repr(numpy.matrix([[1, 2], [3, 4]]),
                                   """\
matrix([[1, 2],
        [3, 4]])""")

        def test_pandas(self):
            # noinspection PyPackageRequirements
            import pandas as pd

            df = pd.DataFrame({'a': [1, 2], 'b': [3, 4]})
            self.assert_usual_repr(df)
            self.assert_usual_repr(df.index)
            self.assert_usual_repr(df.a)
            self.assert_usual_repr(df.b)

            df = pd.DataFrame(
                dict((k, range(100)) for k in 'abcdefghijkl')
            ).set_index(['a', 'b'])

            self.assert_cheap_repr(df,
                                   """\
        c   d   e   f  ...   i   j   k   l
a  b                   ...
0  0    0   0   0   0  ...   0   0   0   0
1  1    1   1   1   1  ...   1   1   1   1
2  2    2   2   2   2  ...   2   2   2   2
3  3    3   3   3   3  ...   3   3   3   3
...    ..  ..  ..  ..  ...  ..  ..  ..  ..
96 96  96  96  96  96  ...  96  96  96  96
97 97  97  97  97  97  ...  97  97  97  97
98 98  98  98  98  98  ...  98  98  98  98
99 99  99  99  99  99  ...  99  99  99  99
[100 rows x 10 columns]""")

            self.assert_cheap_repr(df.c,
                                   """\
a   b
0   0      0
1   1      1
2   2      2
3   3      3
          ..
96  96    96
97  97    97
98  98    98
99  99    99
Name: c, Length: 100, dtype: int64""")

            if version_info[:2] < (3, 6):
                self.assert_cheap_repr(df.index,
                                       """\
MultiIndex(levels=[Int64Index(dtype=dtype('int64'), name='a', length=100), Int64Index(dtype=dtype('int64'), name='b', length=100)],
           codes=[FrozenNDArray([ 0,  1,  2, ..., 97, 98, 99], dtype=int8), FrozenNDArray([ 0,  1,  2, ..., 97, 98, 99], dtype=int8)],
           names=['a', 'b'])""")
            else:
                self.assert_cheap_repr(df.index,
                                       """\
MultiIndex(levels=[Int64Index(dtype=dtype('int64'), length=100), Int64Index(dtype=dtype('int64'), length=100)],
           codes=[array([ 0,  1,  2, ..., 97, 98, 99], dtype=int8), array([ 0,  1,  2, ..., 97, 98, 99], dtype=int8)],
           names=['a', 'b'])""")

            values = [4, 2, 3, 1]
            cats = pd.Categorical([1, 2, 3, 4], categories=values)
            self.assert_cheap_repr(
                pd.DataFrame(
                    {"strings": ["a", "b", "c", "d"], "values": values},
                    index=cats).index,
                "CategoricalIndex(categories=Int64Index(dtype=dtype('int64'), "
                "length=4), ordered=False, dtype='category', length=4)"
            )

            if sys.version_info[:2] == (3, 7):
                expected = """\
IntervalIndex(closed='right',
              dtype=interval[int64, right])"""
            else:
                expected = """\
IntervalIndex(closed='right',
              dtype=interval[int64])"""
            self.assert_cheap_repr(pd.interval_range(start=0, end=5), expected)

    def test_bytes(self):
        self.assert_usual_repr(b'')
        self.assert_usual_repr(b'123')
        self.assert_cheap_repr(b'abc' * 50,
                               "b'abcabcabcabcabcabcabcabcabca...bcabcabcabcabcabcabcabcabcabc'"
                               .lstrip('b' if PY2 else ''))

    def test_str(self):
        self.assert_usual_repr('')
        self.assert_usual_repr(u'')
        self.assert_usual_repr(u'123')
        self.assert_usual_repr('123')
        self.assert_cheap_repr('abc' * 50,
                               "'abcabcabcabcabcabcabcabcabca...bcabcabcabcabcabcabcabcabcabc'")

    def test_inheritance(self):
        class A(object):
            def __init__(self):
                pass

        class B(A):
            pass

        class C(A):
            pass

        class D(C):
            pass

        class C2(C):
            pass

        class C3(C, B):
            pass

        class B2(B, C):
            pass

        class A2(A):
            pass

        @register_repr(A)
        def repr_A(_x, _helper):
            return 'A'

        @register_repr(C)
        def repr_C(_x, _helper):
            return 'C'

        @register_repr(B)
        def repr_B(_x, _helper):
            return 'B'

        @register_repr(D)
        def repr_D(_x, _helper):
            return 'D'

        self.assert_cheap_repr(A(), 'A')
        self.assert_cheap_repr(B(), 'B')
        self.assert_cheap_repr(C(), 'C')
        self.assert_cheap_repr(D(), 'D')
        self.assert_cheap_repr(C2(), 'C')
        self.assert_cheap_repr(C3(), 'C')
        self.assert_cheap_repr(B2(), 'B')
        self.assert_cheap_repr(A2(), 'A')

        self.assertEqual(find_repr_function(A), repr_A)
        self.assertEqual(find_repr_function(B), repr_B)
        self.assertEqual(find_repr_function(C), repr_C)
        self.assertEqual(find_repr_function(D), repr_D)
        self.assertEqual(find_repr_function(C2), repr_C)
        self.assertEqual(find_repr_function(C3), repr_C)
        self.assertEqual(find_repr_function(B2), repr_B)
        self.assertEqual(find_repr_function(A2), repr_A)

    def test_exceptions(self):
        with temp_attrs(cheap_repr, 'raise_exceptions', True):
            self.assertRaises(ValueError,
                              lambda: cheap_repr(ErrorClass(True)))

        for C in [ErrorClass, OldStyleErrorClass]:
            name = C.__name__
            self.assert_usual_repr(C())
            warning_message = "Exception 'ValueError' in repr_object for object of type %s. " \
                              "The repr has been suppressed for this type." % name
            if PY2 and C is OldStyleErrorClass:
                warning_message = warning_message.replace('repr_object', 'repr')
            self.assert_cheap_repr_warns(
                C(True),
                warning_message,
                '<%s instance at 0xXXX (exception in repr)>' % name,
            )
            self.assert_cheap_repr(C(), '<%s instance at 0xXXX (repr suppressed)>' % name)
        for C in [ErrorClassChild, OldStyleErrorClassChild]:
            name = C.__name__
            self.assert_cheap_repr(C(), '<%s instance at 0xXXX (repr suppressed)>' % name)

    def test_func_raise_exceptions(self):
        class T(object):
            pass

        @register_repr(T)
        def bad_repr(*_):
            raise TypeError()

        bad_repr.raise_exceptions = True

        self.assertRaises(TypeError, lambda: cheap_repr(T()))

        class X(object):
            def __repr__(self):
                raise IOError()

        class Y:  # old-style in python 2
            def __repr__(self):
                raise IOError()

        raise_exceptions_from_default_repr()

        for C in [X, Y]:
            self.assertRaises(IOError, lambda: cheap_repr(C()))

    def test_default_too_long(self):
        self.assert_usual_repr(DirectRepr('hello'))
        self.assert_cheap_repr_warns(
            DirectRepr('long' * 500),
            'DirectRepr.__repr__ is too long and has been suppressed. '
            'Register a repr for the class to avoid this warning '
            'and see an informative repr again, '
            'or increase cheap_repr.suppression_threshold',
            'longlonglonglonglonglonglong...glonglonglonglonglonglonglong')
        self.assert_cheap_repr(DirectRepr('hello'),
                               '<DirectRepr instance at 0xXXX (repr suppressed)>')

    def test_maxparts(self):
        self.assert_cheap_repr(list(range(8)),
                               '[0, 1, 2, ..., 5, 6, 7]')
        self.assert_cheap_repr(list(range(20)),
                               '[0, 1, 2, ..., 17, 18, 19]')
        with temp_attrs(find_repr_function(list), 'maxparts', 10):
            self.assert_cheap_repr(list(range(8)),
                                   '[0, 1, 2, 3, 4, 5, 6, 7]')
            self.assert_cheap_repr(list(range(20)),
                                   '[0, 1, 2, 3, 4, ..., 15, 16, 17, 18, 19]')

    def test_recursive(self):
        lst = [1, 2, 3]
        lst.append(lst)
        self.assert_cheap_repr(lst, '[1, 2, 3, [1, 2, 3, [1, 2, 3, [...]]]]')

        d = {1: 2, 3: 4}
        d[5] = d
        self.assert_cheap_repr(
            d, '{1: 2, 3: 4, 5: {1: 2, 3: 4, 5: {1: 2, 3: 4, 5: {...}}}}')

    def test_custom_set(self):
        self.assert_cheap_repr(RangeSet(0), 'RangeSet()')
        self.assert_cheap_repr(RangeSet(3), 'RangeSet({0, 1, 2})')
        self.assert_cheap_repr(RangeSet(10), 'RangeSet({0, 1, 2, 3, 4, 5, ...})')

    def test_repr_function(self):
        def some_really_really_long_function_name():
            yield 3

        self.assert_usual_repr(some_really_really_long_function_name, normalise=True)
        self.assert_usual_repr(some_really_really_long_function_name(), normalise=True)
        self.assert_usual_repr(os)

    @skipIf(PY2 and PYPY, "Not supported in pypy2")
    def test_repr_long_class_name(self):
        class some_really_really_long_class_name(object):
            pass

        self.assert_usual_repr(some_really_really_long_class_name(), normalise=True)

    def test_function_names_unique(self):
        # Duplicate function names can lead to mistakes
        assert_unique(f.__name__ for f in set(repr_registry.values()))

    @skipUnless(PY2, "Old style classes only exist in Python 2")
    def test_old_style_class(self):
        self.assert_cheap_repr(OldStyleClass, '<class tests.utils.OldStyleClass at 0xXXX>')

    def test_target_length(self):
        target = 100
        lst = []
        for i in range(100):
            lst.append(i)
            r = cheap_repr(lst, target_length=target)
            usual = repr(lst)
            assert (
                    (r == usual and len(r) < target) ^
                    ('...' in r and len(r) > target)
            )

        self.assertEqual(
            cheap_repr(list(range(100)), target_length=target),
            '[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, ..., 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99]'
        )

        # Don't want to deal with ordering in older pythons
        if version_info[:2] >= (3, 6):
            self.assertEqual(
                cheap_repr(set(range(100)), target_length=target),
                '{0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, ...}'
            )

            self.assertEqual(
                cheap_repr({x: x * 2 for x in range(100)}, target_length=target),
                '{0: 0, 1: 2, 2: 4, 3: 6, 4: 8, 5: 10, 6: 12, 7: 14, 8: 16, 9: 18, 10: 20, 11: 22, 12: 24, 13: 26, ...}',
            )


if __name__ == '__main__':
    unittest.main()



import traceback

from sys import version_info

try:
    from qualname import qualname
except ImportError:
    def qualname(cls):
        return cls.__qualname__

PY2 = version_info[0] == 2
PY3 = not PY2

if PY2:
    def viewitems(d):
        return d.viewitems()
else:
    def viewitems(d):
        return d.items()


def safe_qualname(cls):
    # type: (type) -> str
    result = _safe_qualname_cache.get(cls)
    if not result:
        try:
            result = qualname(cls)
        except (AttributeError, IOError, SyntaxError):
            result = cls.__name__
        if '<locals>' not in result:
            _safe_qualname_cache[cls] = result
    return result


_safe_qualname_cache = {}


def type_name(x):
    return safe_qualname(x.__class__)


def exception_string(exc):
    assert isinstance(exc, BaseException)
    return ''.join(traceback.format_exception_only(type(exc), exc)).strip()



"""
MIT License

Copyright (c) 2021 Alex Hall

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import inspect
import sys
import warnings
from array import array
from collections import defaultdict, deque
from importlib import import_module
from itertools import islice, repeat

from cheap_repr.utils import type_name, exception_string, safe_qualname, viewitems, PY2, PY3

if PY2:
    from itertools import izip as zip, izip_longest as zip_longest
    from collections import Mapping, Set
else:
    from itertools import zip_longest
    from collections.abc import Mapping, Set

try:
    from .version import __version__
except ImportError:  # pragma: no cover
    # version.py is auto-generated with the git tag when building
    __version__ = "???"


class ReprSuppressedWarning(Warning):
    """
    This warning is raised when a class is supressed from having a
    repr calculated for it by cheap_repr in the future.
    Instead the output will be of the form:
    <MyClass instance at 0x123 (repr suppressed)>
    This can happen when either:
    1. An exception is raised by either repr(x) or f(x) where
       f is a registered repr function for that class. See
       'Exceptions in repr functions' in the README for more.
    2. The output of repr(x) is longer than
       cheap_repr.suppression_threshold characters.
    """


repr_registry = {}


def try_register_repr(module_name, class_name):
    """
    This tries to register a repr function for a class that may not exist,
    e.g. if the class is in a third party package that may not be installed.
    module_name and class_name are strings. If the class can be imported,
    then:

        @try_register_repr(module_name, class_name)
        def repr_function(...)
            ...

    is equivalent to:

        from <module_name> import <class_name>
        @register_repr(<class_name>)
        def repr_function(...)
            ...

    If the class cannot be imported, nothing happens.
    """
    try:
        cls = getattr(import_module(module_name), class_name)
        return register_repr(cls)
    except Exception:
        return lambda x: x


def register_repr(cls):
    """
    Register a repr function for cls. The function must accept two arguments:
    the object to be represented as a string, and an instance of ReprHelper.
    The registered function will be used by cheap_repr when appropriate,
    and can be retrieved by find_repr_function(cls).
    """

    assert inspect.isclass(cls), 'register_repr must be called with a class. ' \
                                 'The type of %s is %s' % (cheap_repr(cls), type_name(cls))

    def decorator(func):
        repr_registry[cls] = func
        func.__dict__.setdefault('maxparts', 6)
        return func

    return decorator


def maxparts(num):
    """
    See the maxparts section in the README.
    """

    def decorator(func):
        func.maxparts = num
        return func

    return decorator


@try_register_repr('pandas.core.internals', 'BlockManager')
def basic_repr(x, *_):
    return '<%s instance at %#x>' % (type_name(x), id(x))


@try_register_repr('importlib.machinery', 'ModuleSpec')
@register_repr(type(register_repr))
@register_repr(type(_ for _ in []))
@register_repr(type(inspect))
def normal_repr(x, *_):
    """
    Register this with a class to indicate that its own
    __repr__ method is already fine. This prevents it from
    being supressed when its output is a bit long.
    """
    return repr(x)


suppressed_classes = set()


@register_repr(object)
@maxparts(60)
def repr_object(x, helper):
    s = repr(x)
    if type(x).__repr__ is object.__repr__:
        return s

    if len(s) > cheap_repr.suppression_threshold:
        cls = x.__class__
        suppressed_classes.add(cls)
        warnings.warn(ReprSuppressedWarning(
            '%s.__repr__ is too long and has been suppressed. '
            'Register a repr for the class to avoid this warning '
            'and see an informative repr again, '
            'or increase cheap_repr.suppression_threshold' % safe_qualname(cls)))
    return helper.truncate(s)


def find_repr_function(cls):
    for cls in inspect.getmro(cls):
        func = repr_registry.get(cls)
        if func:
            return func


__raise_exceptions_from_default_repr = False


def raise_exceptions_from_default_repr():
    global __raise_exceptions_from_default_repr
    __raise_exceptions_from_default_repr = True
    repr_object.raise_exceptions = True


def cheap_repr(x, level=None, target_length=None):
    """
    Return a short, computationally inexpensive string
    representation of x, with approximately up to `level`
    levels of nesting.
    """
    if level is None:
        level = cheap_repr.max_level
    x_cls = getattr(x, '__class__', type(x))
    for cls in inspect.getmro(x_cls):
        if cls in suppressed_classes:
            return _basic_but('repr suppressed', x)
        func = repr_registry.get(cls)
        if func:
            helper = ReprHelper(level, func, target_length)
            return _try_repr(func, x, helper)

    # Old-style classes in Python 2.
    return _try_repr(repr, x)


cheap_repr.suppression_threshold = 300
cheap_repr.raise_exceptions = False
cheap_repr.max_level = 3


def _try_repr(func, x, *args):
    try:
        return func(x, *args)
    except BaseException as e:
        should_raise = (cheap_repr.raise_exceptions or
                        getattr(func, 'raise_exceptions', False) or
                        func is repr and __raise_exceptions_from_default_repr)
        if should_raise:
            raise
        cls = x.__class__
        if cls not in suppressed_classes:
            suppressed_classes.add(cls)
            warnings.warn(ReprSuppressedWarning(
                "Exception '%s' in %s for object of type %s. "
                "The repr has been suppressed for this type." %
                (exception_string(e), func.__name__, safe_qualname(cls))))
        return _basic_but('exception in repr', x)


def _basic_but(message, x):
    return '%s (%s)>' % (basic_repr(x)[:-1], message)


class ReprHelper(object):
    __slots__ = ('level', 'func', 'target_length')

    def __init__(self, level, func, target_length):
        self.level = level
        self.func = func
        self.target_length = target_length

    def repr_iterable(self, iterable, left, right, end=False, length=None):
        """
        Produces a comma-separated representation of `iterable`, automatically handling nesting and iterables
        that are too long, surrounded by `left` and `right`.
        The number of items is at most `maxparts` (see the configuration section in the README).

        Set `end=True` to include items from both the beginning and end, possibly leaving out items
        in the middle. Only do this if `iterable` supports efficient slicing at the end, e.g. `iterable[-3:]`.

        Provide the `length` parameter if `len(iterable)` doesn't work. Usually this is not needed.
        """

        if length is None:
            length = len(iterable)
        if self.level <= 0 and length:
            s = '...'
        else:
            newlevel = self.level - 1
            max_parts = original_maxparts = self.func.maxparts
            truncate = length > max_parts
            target_length = self.target_length

            if target_length is None or not truncate:
                if end and truncate:
                    # Round up from half, e.g. 7 -> 4
                    max_parts -= max_parts // 2

                pieces = [cheap_repr(elem, newlevel) for elem in islice(iterable, max_parts)]

                if truncate:
                    pieces.append('...')

                    if end:
                        max_parts = original_maxparts - max_parts

                        pieces += [cheap_repr(elem, newlevel) for elem in iterable[-max_parts:]]

            else:
                pieces = []
                right_pieces = []
                total_length = 3 + len(left) + len(right)

                if end:
                    sentinel = object()

                    def parts_gen():
                        for left_part, right_part in zip_longest(
                                iterable[:length - length // 2],
                                islice(reversed(iterable), length // 2),
                                fillvalue=sentinel,
                        ):
                            yield left_part, pieces
                            if right_part is not sentinel:
                                yield right_part, right_pieces

                    parts = parts_gen()
                else:
                    parts = zip(iterable, repeat(pieces))

                for i, (part, pieces_list) in enumerate(parts):
                    r = cheap_repr(part, newlevel)
                    pieces_list.append(r)
                    total_length += len(r) + 2
                    if max_parts <= i + 1 < length and total_length >= target_length:
                        pieces_list.append('...')
                        break

                if end:
                    pieces += right_pieces[::-1]

            s = ', '.join(pieces)
        return left + s + right

    def truncate(self, string, middle='...'):
        """
        Returns a version of `string` at most `maxparts` characters long,
        with the middle replaced by `...` if necessary.
        """
        max_parts = self.func.maxparts
        if len(string) > max_parts:
            i = max(0, (max_parts - 3) // 2)
            j = max(0, max_parts - 3 - i)
            string = string[:i] + middle + string[len(string) - j:]
        return string


@register_repr(type(ReprHelper(0, None, None).truncate))
def repr_bound_method(meth, _helper):
    obj = meth.__self__
    return '<bound method %s.%s of %s>' % (
        type_name(obj), meth.__name__, cheap_repr(obj))


@register_repr(tuple)
def repr_tuple(x, helper):
    if len(x) == 1:
        return '(%s,)' % cheap_repr(x[0], helper.level)
    else:
        return helper.repr_iterable(x, '(', ')', end=True)


@register_repr(list)
@try_register_repr('UserList', 'UserList')
@try_register_repr('collections', 'UserList')
def repr_list(x, helper):
    return helper.repr_iterable(x, '[', ']', end=True)


@register_repr(array)
@maxparts(5)
def repr_array(x, helper):
    if not x:
        return repr(x)
    return helper.repr_iterable(x, "array('%s', [" % x.typecode, '])', end=True)


@register_repr(set)
def repr_set(x, helper):
    if not x:
        return repr(x)
    elif PY2:
        return helper.repr_iterable(x, 'set([', '])')
    else:
        return helper.repr_iterable(x, '{', '}')


@register_repr(frozenset)
def repr_frozenset(x, helper):
    if not x:
        return repr(x)
    elif PY2:
        return helper.repr_iterable(x, 'frozenset([', '])')
    else:
        return helper.repr_iterable(x, 'frozenset({', '})')


@register_repr(Set)
def repr_Set(x, helper):
    if not x:
        return '%s()' % type_name(x)
    else:
        return helper.repr_iterable(x, '%s({' % type_name(x), '})')


@register_repr(deque)
def repr_deque(x, helper):
    return helper.repr_iterable(x, 'deque([', '])')


class _DirectRepr(str):
    def __repr__(self):
        return self


register_repr(_DirectRepr)(normal_repr)


@register_repr(dict)
@maxparts(4)
def repr_dict(x, helper):
    newlevel = helper.level - 1

    return helper.repr_iterable(
        (
            _DirectRepr('%s: %s' % (
                cheap_repr(key, newlevel),
                cheap_repr(x[key], newlevel),
            ))
            for key in x
        ),
        '{', '}', length=len(x),
    )


@try_register_repr('__builtin__', 'unicode')
@register_repr(str)
@maxparts(60)
def repr_str(x, helper):
    return repr(helper.truncate(x))


if PY3:
    @register_repr(bytes)
    @maxparts(60)
    def repr_bytes(x, helper):
        return repr(helper.truncate(x, middle=b'...'))


@register_repr(int)
@try_register_repr('__builtin__', 'long')
@maxparts(40)
def repr_int(x, helper):
    return helper.truncate(repr(x))


@try_register_repr('numpy', 'ndarray')
def repr_ndarray(x, _helper):
    # noinspection PyPackageRequirements
    import numpy as np

    dims = len(x.shape)
    if (
            # Too many dimensions to be concise
            dims > 6 or
            # There's a bug with array_repr and matrices
            isinstance(x, np.matrix) and np.lib.NumpyVersion(np.__version__) < '1.14.0' or
            # and with masked arrays...
            isinstance(x, np.ma.MaskedArray)

    ):
        name = type_name(x)
        if name == 'ndarray':
            name = 'array'
        return '%s(%r, shape=%r)' % (name, x.dtype, x.shape)

    edgeitems = repr_ndarray.maxparts // 2
    if dims == 3:
        edgeitems = min(edgeitems, 2)
    elif dims > 3:
        edgeitems = 1

    opts = np.get_printoptions()
    try:
        np.set_printoptions(threshold=repr_ndarray.maxparts, edgeitems=edgeitems)
        return np.array_repr(x)
    finally:
        np.set_printoptions(**opts)


@try_register_repr('pandas', 'DataFrame')
def repr_DataFrame(df, _):
    """
    This function can be configured by setting `max_rows` or `max_cols` attributes.
    """
    from pandas import get_option

    return df.to_string(
        max_rows=repr_DataFrame.max_rows,
        max_cols=repr_DataFrame.max_cols,
        show_dimensions=get_option("display.show_dimensions"),
    )


repr_DataFrame.max_rows = 8
repr_DataFrame.max_cols = 8


@try_register_repr('pandas', 'Series')
def repr_pandas_Series(series, _):
    """
    This function can be configured by setting the `max_rows` attributes.
    """
    from pandas import get_option

    return series.to_string(
        max_rows=repr_pandas_Series.max_rows,
        name=series.name,
        dtype=series.dtype,
        length=get_option("display.show_dimensions"),
    )


repr_pandas_Series.max_rows = 8


def _repr_pandas_index_generic(index, helper, attrs, long_space=False):
    klass = index.__class__.__name__
    if long_space:
        space = '\n%s' % (' ' * (len(klass) + 1))
    else:
        space = ' '

    prepr = (",%s" % space).join(
        "%s=%s" % (k, cheap_repr(v, helper.level - 1))
        for k, v in attrs)
    return "%s(%s)" % (klass, prepr)


@try_register_repr('pandas', 'Index')
def repr_pandas_Index(index, helper):
    attrs = [['dtype', index.dtype]]
    if index.name is not None:
        attrs.append(['name', index.name])
    attrs.append(['length', len(index)])
    return _repr_pandas_index_generic(index, helper, attrs)


@try_register_repr('pandas', 'IntervalIndex')
def repr_pandas_IntervalIndex(index, helper):
    attrs = [['closed', index.closed]]
    if index.name is not None:
        attrs.append(['name', index.name])
    attrs.append(['dtype', index.dtype])
    return _repr_pandas_index_generic(index, helper, attrs, long_space=True)


@try_register_repr('pandas', 'RangeIndex')
def repr_pandas_RangeIndex(index, helper):
    attrs = index._format_attrs()
    return _repr_pandas_index_generic(index, helper, attrs)


@try_register_repr('pandas', 'MultiIndex')
def repr_pandas_MultiIndex(index, helper):
    attrs = [('levels', index._levels)]

    try:
        attrs.append(('labels', index._labels))
    except AttributeError:
        attrs.append(('codes', index.codes))

    attrs.append(('names', index.names))

    if index.sortorder is not None:
        attrs.append(('sortorder', index.sortorder))
    return _repr_pandas_index_generic(index, helper, attrs, long_space=True)


@try_register_repr('pandas', 'CategoricalIndex')
def repr_pandas_CategoricalIndex(index, helper):
    attrs = [('categories', index.categories),
             ('ordered', index.ordered)]
    if index.name is not None:
        attrs.append(['name', index.name])
    attrs.append(['dtype', index.dtype.name])
    attrs.append(['length', len(index)])
    return _repr_pandas_index_generic(index, helper, attrs)


@try_register_repr('django.db.models', 'QuerySet')
def repr_QuerySet(x, _):
    try:
        model_name = x.model._meta.object_name
    except AttributeError:
        model_name = type_name(x.model)
    return '<%s instance of %s at %#x>' % (type_name(x), model_name, id(x))


@try_register_repr('collections', 'ChainMap')
@try_register_repr('chainmap', 'ChainMap')
@maxparts(4)
def repr_ChainMap(x, helper):
    return helper.repr_iterable(x.maps, type_name(x) + '(', ')', end=True)


def _register_ordered_dict(f):
    f = try_register_repr('collections', 'OrderedDict')(f)
    f = try_register_repr('ordereddict', 'OrderedDict')(f)
    f = try_register_repr('backport_collections', 'OrderedDict')(f)
    return f

@try_register_repr('collections', 'UserDict')
@try_register_repr('UserDict', 'UserDict')
@register_repr(Mapping)
@maxparts(5)
def repr_Mapping(x, helper):
    if not x:
        return type_name(x) + '()'
    return '{0}({1})'.format(type_name(x), repr_dict(x, helper))


if sys.version_info < (3, 12):
    @maxparts(4)
    def repr_OrderedDict(x, helper):
        if not x:
            return repr(x)
        helper.level += 1
        return helper.repr_iterable(viewitems(x), type_name(x) + '([', '])', length=len(x))


    _register_ordered_dict(repr_OrderedDict)
else:
    _register_ordered_dict(repr_Mapping)

@try_register_repr('collections', 'Counter')
@try_register_repr('counter', 'Counter')
@try_register_repr('backport_collections', 'Counter')
@maxparts(5)
def repr_Counter(x, helper):
    length = len(x)
    if length <= repr_Counter.maxparts:
        return repr_Mapping(x, helper)
    else:
        # The default repr of Counter gives the items in decreasing order
        # of frequency. We don't do that because it would be expensive
        # to compute. We also don't show a sample of random keys
        # because someone familiar with the default repr might be misled
        # into thinking that they are the most common.
        return '{0}({1} keys)'.format(type_name(x), length)


@register_repr(defaultdict)
@maxparts(4)
def repr_defaultdict(x, helper):
    return '{0}({1}, {2})'.format(type_name(x),
                                  x.default_factory,
                                  repr_dict(x, helper))


def repr_Printer(x, _helper):
    contents = repr(x)
    return '{0}({1})'.format(type_name(x),
                             cheap_repr(contents))


try:
    register_repr(type(copyright))
except NameError:
    pass

if PY3:
    @register_repr(type({}.keys()))
    def repr_dict_keys(x, helper):
        return helper.repr_iterable(x, 'dict_keys([', '])')


    @register_repr(type({}.values()))
    def repr_dict_values(x, helper):
        return helper.repr_iterable(x, 'dict_values([', '])')


    @register_repr(type({}.items()))
    @maxparts(4)
    def repr_dict_items(x, helper):
        helper.level += 1
        return helper.repr_iterable(x, 'dict_items([', '])')
